#include "app/framework/include/af.h"
#include "app/util/zigbee-framework/zigbee-device-common.h"
#include "stdio.h"
#include "stdbool.h"

#include "ikk-debug.h"
#include "ikk-module-def.h"
#include "ikk-common-utils.h"
#include "ikk-opt-tunnel.h"
#include "ikk-command.h"
#include "ikk-network.h"


// command incoming global manager.
CommandIncomingMagSt g_stCmdIncomingMager = { 0 };
// command outgoing global manager.
CommandOutgoingMagSt g_stCmdOutgoingMager = { 0 };

// supported commands list
CommandDefSt g_stCommandsDef[] = STATIC_COMMAND_DEF;

// command images cache list for ack reply detection
CommandImageNodeSt g_stCmdImageList[CMDIMAGE_LIST_DEPTH] = { 0 };

EmberEventControl kCmdCommandImageHandleEventControl;

// link layer ack enable switch. when disable, the downlink command will no longer check errors and confirm
// , and the uplink command will no longer be re-send!!!
static bool g_bLinkAckEnableFlag = true;
static bool g_bLinkCRCVerifyEnableFlag = true;

// EXTERN FUNCTIONS DECLARE
void emberAfGetEui64(EmberEUI64 returnEui64);
EmberNodeId emberAfGetNodeId(void);
EmberStatus emberAfGetNetworkParameters(EmberNodeType* nodeType, EmberNetworkParameters* parameters);
EmberStatus emberAfPluginNetworkCreatorNetworkForm(bool centralizedNetwork, EmberPanId panId, int8_t radioTxPower, uint8_t channel);
void removeServer(uint8_t* ieeeAddress);
uint8_t emberAfGetChildTableSize(void);
void networkLeaveCommand(void);
void NWK_ClearRssi(void);
EmberStatus emberLeaveRequest(EmberNodeId target, EmberEUI64 deviceAddress, uint8_t leaveRequestFlags, EmberApsOption options);
//void emAfClearServers(void);
EmberStatus emberAfGetChildData(uint8_t index, EmberChildData *childData);
//EmberStatus emberAfPluginNetworkCreatorSecurityOpenNetwork(void);
//EmberStatus emberAfPluginNetworkCreatorSecurityCloseNetwork(void);
//EmberStatus emberAfPluginNetworkCreatorStart(bool centralizedNetwork);
void emAfApsFrameClusterIdSetup(uint16_t clusterId);
EmberStatus emberAfSendUnicast(EmberOutgoingMessageType type, uint16_t indexOrDestination, EmberApsFrame *apsFrame,
                               uint16_t messageLength, uint8_t *message);
bool emberAfOtaServerSendImageNotifyCallback(EmberNodeId dest, uint8_t endpoint, uint8_t payloadType, uint8_t queryJitter, const EmberAfOtaImageId* id);
//bool kkOtaImageBlockWrite(int offset_in, char *pdata );


// Obtain the next sequence number to be used in ZDP message, application layer!!!
#define ZDPAPP_NEXT_SEQ() ((emberGetLastAppZigDevRequestSequence() + 1) & 0X7F/*SAME TO APPLICATION_ZDO_SEQUENCE_MASK*/)
// Obtain the next sequence number to be used in ZDP message, stack layer!!!
#define ZDPSTK_NEXT_SEQ() ((emberGetLastStackZigDevRequestSequence() == 0XFF)?(0X80):(emberGetLastStackZigDevRequestSequence() + 1))
// Obtain the next sequence number to be used in ZCL message, application layer!!!
#define ZCLAPP_NEXT_SEQ() ((emberAfGetLastSequenceNumber() + 1) & 0X7F/*SAME TO EMBER_AF_ZCL_SEQUENCE_MASK*/)


/*	DESP: get the corresponding command[req] name according to the command type.
	Auth: dingmz_frc.20191019.
*/
char *kCmdGetCommandName(CommandEnum ecommand )
{
	CommandEnum mapp_command = ((SECTION_ID(ecommand) == DCMD_ACK_SECTION_ID) || (SECTION_ID(ecommand) == UCMD_ACK_SECTION_ID))?
		(ecommand - CMD_REQ_ACK_DIFF):ecommand;

	for(int index = 0; index < (sizeof(g_stCommandsDef) / sizeof(CommandDefSt)); ++index ) {
		if( g_stCommandsDef[index].command == mapp_command ) {
			if((SECTION_ID(ecommand) == DCMD_ACK_SECTION_ID) || (SECTION_ID(ecommand) == UCMD_ACK_SECTION_ID)) {
				return g_stCommandsDef[index].cmdname_ack;
			}else {
				return g_stCommandsDef[index].cmdname_req;
			}
		}
	}

	return NULL;
}

/*	DESP: get the corresponding command type according to the command name.
	Auth: dingmz_frc.20191019.
*/
CommandEnum kCmdGetCommandID(char *command_name )
{
	if( NULL == command_name ) {
		return ECMD_UNKNOW;
	}

	for(int index = 0; index < (sizeof(g_stCommandsDef) / sizeof(CommandDefSt)); ++index ) {
		if( S_CMP(g_stCommandsDef[index].cmdname_req, command_name)) {
			return g_stCommandsDef[index].command;
		}else if( S_CMP(g_stCommandsDef[index].cmdname_ack, command_name)) {
			return CMDACK(g_stCommandsDef[index].command);
		}
	}

	return ECMD_UNKNOW;
}

/*	DESP: get the corresponding command type according to the command ID.
	Auth: dingmz_frc.20191019.
*/
CommandTypeEnum kCmdGetCommandType(CommandEnum ecommand )
{
	for(int index = 0; index < (sizeof(g_stCommandsDef) / sizeof(CommandDefSt)); ++index ) {
		if( g_stCommandsDef[index].command == ecommand ) {
			return g_stCommandsDef[index].type;
		}
	}

	return NULL;
}

/*	DESP: get the corresponding command pre-configuration record according to the command type.
	Auth: dingmz_frc.20191019.
*/
CommandDefSt *kCmdGetCommandConf(CommandEnum ecommand )
{
	for(int index = 0; index < (sizeof(g_stCommandsDef) / sizeof(CommandDefSt)); ++index ) {
		if( g_stCommandsDef[index].command == ecommand ) {
			return &g_stCommandsDef[index];
		}
	}

	return NULL;
}

/*	DESP: add one command record to the image list.
	Auth: dingmz_frc.20191101.
*/
kk_err_t kCmdAddCommandRecord2ImageList(CommandImageNodeSt *command_image )
{
	if( NULL == command_image ) {
		return KET_ERR_INVALID_POINTER;
	}

	#if 0
	iKonkeAfSelfPrint("ImageAdd: cmd(%d), node_id(%2x), ep(%d), cluster(%2x),cls_cmd(%x), cmd_seq(%d)\r\n"
		, command_image->command, command_image->node_id, command_image->endpoint
		, command_image->cluster_id, command_image->cls_cmd_id, command_image->command_seq);
	#endif

	int valid_index = CMDIMAGE_LIST_DEPTH;
	kk_err_t err = KET_OK;

	for(int index = 0; index < CMDIMAGE_LIST_DEPTH; ++index) {
		if( M_CMP(&g_stCmdImageList[index], command_image, sizeof(CommandImageNodeSt))) {
			valid_index = index;
			break;
		}else if((g_stCmdImageList[index].command == ECMD_UNKNOW) && (valid_index == CMDIMAGE_LIST_DEPTH)) {
			valid_index = index;
		}
	}

	if( valid_index < CMDIMAGE_LIST_DEPTH ) {
		memcpy(&g_stCmdImageList[valid_index], command_image, sizeof(CommandImageNodeSt));
	}else {
		err = KET_ERR_NO_PERMISSIONS;
	}

	return err;
}

/*	DESP: Remove up-request command records from image list by command sequence.
	Auth: dingmz_frc.20191102.
*/
kk_err_t kCmdRemoveOutgoingCommandRecordsBySeq(uint8_t up_seq )
{
	kk_err_t err = KET_OK;

	for(int index = 0; index < CMDIMAGE_LIST_DEPTH; ++index) {
		uint8_t section_id = SECTION_ID(g_stCmdImageList[index].command);

		if(	(section_id == DCMD_ACK_SECTION_ID || section_id == UCMD_REQ_SECTION_ID) &&
			(g_stCmdImageList[index].command_seq == up_seq)) {
			memset(&g_stCmdImageList[index], 0, sizeof(CommandImageNodeSt));
		}
	}

	return err;
}

/*	DESP: put in an 8-bit variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInU8int(uint8_t param_v, bool bend )
{
	char oct_char[8] = { 0 };

	sprintf(oct_char, "%02x%s", param_v, bend?":":"");
	strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);

	return KET_OK;
}

/*	DESP: put in an 16-bit variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInU16int(uint16_t param_v, bool bend )
{
	char oct_char[8] = { 0 };

	sprintf(oct_char, "%04x%s", param_v, bend?":":"");
	strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);

	return KET_OK;
}

/*	DESP: put in an 32-bit variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInU32int(uint32_t param_v, bool bend )
{
	char oct_char[16] = { 0 };

	sprintf(oct_char, "%08x%s", param_v, bend?":":"");
	strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);

	return KET_OK;
}

/*	DESP: put in an 32-bit variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInU64int(uint64_t param_v, bool bend )
{
	char oct_char[24] = { 0 };

	sprintf(oct_char, "%016llx%s", param_v, bend?":":"");
	strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);

	return KET_OK;
}

/*	DESP: put in an 8-bit variable array to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInU8Array(uint8_t param_v[], int length, bool bend )
{
	char oct_char[3] = { 0 };

	for(int index = 0; index < length; ++index ) {
		sprintf(oct_char, "%02x", param_v[index]);
		strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);
	}

	if( bend ) {
		strncat(g_stCmdOutgoingMager.cmdpmls, ":", COMMAND_PARAM_MAXLEN);
	}

	return KET_OK;
}

/*	DESP: put in an int variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInInt(int param_v, bool bend )
{
	char oct_char[16] = { 0 };

	sprintf(oct_char, "%d%s", param_v, bend?":":"");
	strncat(g_stCmdOutgoingMager.cmdpmls, oct_char, COMMAND_PARAM_MAXLEN);

	return KET_OK;
}

/*	DESP: put in an octet-string variable to the command parameter list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandPutInOctetString(char *param_v, bool bend )
{
	strncat(g_stCmdOutgoingMager.cmdpmls, param_v, COMMAND_PARAM_MAXLEN);

	if( bend ) {
		strncat(g_stCmdOutgoingMager.cmdpmls, ":", COMMAND_PARAM_MAXLEN);
	}

	return KET_OK;
}

/*	DESP: outgoing sending command setup initialization settings.
 * 	Note: In this interface, it is necessary to check whether there is a corresponding request command image. If there is one,
 * 	the current outgoing command is the response of this image.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandSetUp(CommandEnum command, uint16_t node_id, uint8_t endpoint, uint16_t cluster_id
	, uint8_t cls_cmd_id, uint8_t transmit_seq )
{
	if( ECMD_UNKNOW == command ) {
		return KET_ERR_INVALID_PARAM;
	}

	#if 0
	iKonkeAfSelfPrint("OutgoingCommandSetUp: command(%s), node(%2x), ep(%d), cluster(%2x), cls_cmd_id(%x), trans_seq(%d).\r\n"
		, kCmdGetCommandName(command), node_id, endpoint, cluster_id, cls_cmd_id, transmit_seq);
	#endif

	kk_err_t err = KET_ERR_OPRATE_ILLEGAL;

	strncpy(g_stCmdOutgoingMager.cmdname, kCmdGetCommandName(command), COMMAND_NAME_MAXLEN);
	memset(g_stCmdOutgoingMager.cmdpmls, 0, COMMAND_PARAM_MAXLEN);

	if((command == ECMD_ULINK_ACK) && (g_stCmdIncomingMager.command != ECMD_DLINK_ACK)) {
		g_stCmdOutgoingMager.seq = g_stCmdIncomingMager.seq;
		return KET_OK;
	}

	// outgoing for down request command reply ack, command image must be exists!!!
	if( SECTION_ID(command) == DCMD_ACK_SECTION_ID ) {
		// Check whether the corresponding command image exists
		for(int index = 0; index < CMDIMAGE_LIST_DEPTH; ++index ) {
			#if 0
			iKonkeAfSelfPrint("I-ITEM[%02d] command(%d), node(%2x), ep(%d), cluster(%2x), cls_cmd_id(%x), attr_id(%2x)\r\n", index
				, g_stCmdImageList[index].command, g_stCmdImageList[index].node_id, g_stCmdImageList[index].endpoint
				, g_stCmdImageList[index].cluster_id, g_stCmdImageList[index].cls_cmd_id, g_stCmdImageList[index].attr_id);
			#endif
			if(	(g_stCmdImageList[index].command == command) && (g_stCmdImageList[index].cluster_id == cluster_id) &&
				(g_stCmdImageList[index].node_id == 0xffff || g_stCmdImageList[index].node_id == node_id) &&
				(g_stCmdImageList[index].endpoint == 255 || g_stCmdImageList[index].endpoint == endpoint) &&
				(g_stCmdImageList[index].cls_cmd_id == cls_cmd_id) && (g_stCmdImageList[index].transmit_seq == transmit_seq)) {
				// Match successfully, use the command ack name.
				strncpy(g_stCmdOutgoingMager.cmdname, kCmdGetCommandName(command), COMMAND_NAME_MAXLEN);
				// Match successfully, use the command seq record.
				g_stCmdOutgoingMager.seq = g_stCmdImageList[index].command_seq;

				if((g_stCmdImageList[index].node_id != 0xffff) && (g_stCmdImageList[index].endpoint != 255)) {
					memset(&g_stCmdImageList[index], 0, sizeof(CommandImageNodeSt));
				}

				err = KET_OK;
				break;
			}
		}
	}
	// up-request or notify report command
	else if((SECTION_ID(command) == UCMD_REQ_SECTION_ID) || (SECTION_ID(command) == UCMD_NOTIFY_SECTION_ID)) {
		g_stCmdOutgoingMager.seq = g_stCmdOutgoingMager.seq_counter++;
		err = KET_OK;
	}

	return err;
}

/*	DESP: outgoing command sending interface.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdOutgoingCommandSent(void )
{
	// append the sequence to param list.
	kCmdOutgoingCommandPutInU8int(g_stCmdOutgoingMager.seq, true);

	// calculate the crc16 value of param list.
	uint16_t crc16 = kUtilsCRC16Sum((unsigned char *)g_stCmdOutgoingMager.cmdpmls, strlen(g_stCmdOutgoingMager.cmdpmls));
	// append the crc to param list.
	kCmdOutgoingCommandPutInU16int(crc16, true);

	g_stCmdOutgoingMager.cmdpmls[strlen(g_stCmdOutgoingMager.cmdpmls) - 1] = '\0';

	CMDSTR_SENT_NL("%s{%s}", g_stCmdOutgoingMager.cmdname, g_stCmdOutgoingMager.cmdpmls);

	// outgoing message re-send for up commands
	uint16_t command_id = kCmdGetCommandID(g_stCmdOutgoingMager.cmdname);
	uint8_t  section_id = SECTION_ID(command_id);

	if((g_bLinkAckEnableFlag) && (section_id == DCMD_ACK_SECTION_ID || section_id == UCMD_REQ_SECTION_ID)) {
		if( command_id != ECMD_ULINK_ACK ) {
			CommandImageNodeSt command_image = { ECMD_UNKNOW, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW
				, TRANSMIT_SEQ_UNKNOW, 0, 0, 0 };

			command_image.command = command_id;
			strncpy(command_image.cmdpmls, g_stCmdOutgoingMager.cmdpmls, sizeof(command_image.cmdpmls));
			command_image.command_seq = g_stCmdOutgoingMager.seq;
			command_image.valid_counter = UCMD_RETRY_PERIOD / IMAGE_LOOP_NMS;
			command_image.retry_counter = UCMD_RETRY_COUNT;

			kCmdAddCommandRecord2ImageList(&command_image);
		}
	}

	return KET_OK;
}

/*	DESP: Generate image according to the current incoming command and add it to the image list.
	Auth: dingmz_frc.20191019.
*/
kk_err_t kCmdGenerateIncomingCommandImage(void )
{
	CommandImageNodeSt image_node = { ECMD_UNKNOW, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW
		, TRANSMIT_SEQ_UNKNOW, 0, 0, 0 };

	if( g_stCmdIncomingMager.command == ECMD_DLINK_ACK ) {
		return KET_OK;
	}

	kk_err_t err = KET_OK;

	CommandTypeEnum cmd_t = kCmdGetCommandType(g_stCmdIncomingMager.command);

	image_node.valid_counter = COMMAND_IMAGE_DEADLINE / IMAGE_LOOP_NMS;

	switch(g_stCmdIncomingMager.command) {
		case (ECMD_CLUSTER_CMD_REQ):
		case (ECMD_READ_ATTR_REQ):
		case (ECMD_WRITE_ATTR_REQ):
		{
			image_node.node_id = (uint16_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);

			#if 0 // Broadcast message does not wait for reply response!
			if((image_node.node_id == 0XFFFC) || (image_node.node_id == 0XFFFD) || (image_node.node_id == 0XFFFF)) {
				return KET_OK;
			}
			#endif

			image_node.endpoint = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			image_node.cluster_id = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2], 4);

			if( g_stCmdIncomingMager.command == ECMD_READ_ATTR_REQ ) {
				image_node.cls_cmd_id = 0X01;
			}else if( g_stCmdIncomingMager.command == ECMD_WRITE_ATTR_REQ ) {
				image_node.cls_cmd_id = 0X04;
			}

			image_node.transmit_seq = ZCLAPP_NEXT_SEQ();
			image_node.command = g_stCmdIncomingMager.command + CMD_REQ_ACK_DIFF;
			image_node.command_seq = g_stCmdIncomingMager.seq;
			break;
		}
		// bind request needs to be judged by the ZDO sequence number as the response, so needs to be recorded here!!!
		case (ECMD_LEAVE_REQ):
		case (ECMD_BIND_REQ):
		case (ECMD_ACTIVE_EPS_REQ):
		case (ECMD_SIMPLE_DESP_REQ):
		case (ECMD_MATCH_DESP_REQ):
		case (ECMD_NODE_DESP_REQ):
		case (ECMD_OOD_CMD_REQ):
		{
			image_node.node_id = (uint16_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);

			#if 0 // Broadcast message does not wait for reply response!
			if((image_node.node_id == 0XFFFC) || (image_node.node_id == 0XFFFD) || (image_node.node_id == 0XFFFF)) {
				return KET_OK;
			}
			#endif

			// ECMD_IEEE_ADDR_REQ and ECMD_IP_ADDR_REQ is using stack sequence
			image_node.transmit_seq = (cmd_t == ECT_ZCL)?ZCLAPP_NEXT_SEQ():ZDPAPP_NEXT_SEQ();
			image_node.command = g_stCmdIncomingMager.command + CMD_REQ_ACK_DIFF;
			image_node.command_seq = g_stCmdIncomingMager.seq;
			break;
		}
		case (ECMD_IEEE_ADDR_REQ):
		case (ECMD_IP_ADDR_REQ):
		{
			// ECMD_IEEE_ADDR_REQ and ECMD_IP_ADDR_REQ is using stack sequence
			image_node.transmit_seq = ZDPSTK_NEXT_SEQ();
			image_node.command = g_stCmdIncomingMager.command + CMD_REQ_ACK_DIFF;
			image_node.command_seq = g_stCmdIncomingMager.seq;
			break;
		}
		// Other local operations do not need to be added to the image list!
		default:
		{
			image_node.command = g_stCmdIncomingMager.command + CMD_REQ_ACK_DIFF;
			image_node.command_seq = g_stCmdIncomingMager.seq;
			break;
		}
	}

	// need to add to the image list.
	if( image_node.command != ECMD_UNKNOW ) {
		err = kCmdAddCommandRecord2ImageList(&image_node);
	}

	return err;
}

/*	DESP: separate command message to segments[key, value].
	Auth: dingmz_frc.20180831.
*/
static kk_err_t kCmdCommandIncomingSegmentSplit(void )
{
	if( g_stCmdIncomingMager.counter <= 0 ) {
		return KET_ERR_CMD_FORMAT;
	}

#define _SPE_S_CONT		">{"

	kk_err_t err = KET_OK;

	//_TRIM_STRING_BOTHENDS(msg_content, TRIM_BOTH);
	g_stCmdIncomingMager.seq = 0;

	char *pparse = strstr(g_stCmdIncomingMager.cmdlins, _SPE_S_CONT);

	// Check whether the instruction format and length are correct!!!
	if( pparse ) {
		// delete the last parameter delimiter.
		g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter - 1] = '\0';
		*(++pparse) = '\0';

		//iKonkeAfSelfPrint(";;key: %s\r\n", g_stCmdIncomingMager.cmdlins);
		//iKonkeAfSelfPrint(";;val: %s\r\n", pparse + 1);

		pparse++;

		// calculate parameter list length, including sequence and crc16 value.
		int params_list_length = g_stCmdIncomingMager.counter - (pparse - g_stCmdIncomingMager.cmdlins) - 1;
		// calculate the original parameter list crc16 value before spliting it!
		uint16_t orgl_crc16 = kUtilsCRC16Sum((uint8_t *)pparse, params_list_length - 4);

		// split command params
		char *pparam = strtok(pparse, ":");
		int pm_counter = 0;

		memset(g_stCmdIncomingMager.param, NULL, COMMAND_PARAM_MAXN);
		while(pparam) {
			g_stCmdIncomingMager.param[pm_counter++] = pparam;
			//iKonkeAfSelfPrint("PM[%d]: %s\r\n", pm_counter - 1, g_stCmdIncomingMager.param[pm_counter - 1]);
			pparam = strtok(NULL, ":");
		}

		if( pm_counter >= 2 ) {
			// check the validity of the sequence number.
			err = (strlen(g_stCmdIncomingMager.param[pm_counter - 2]) == 2)?err:KET_ERR_INVALID_PARAM;
			// check the validity of the crc16 value.
			err = (strlen(g_stCmdIncomingMager.param[pm_counter - 1]) == 4)?err:KET_ERR_INVALID_PARAM;
		}else {
			err = KET_ERR_CMD_FORMAT;
		}

		if( err == KET_OK ) {
			// split command name
			g_stCmdIncomingMager.command = kCmdGetCommandID(g_stCmdIncomingMager.cmdlins);

			// get seq
			g_stCmdIncomingMager.seq   = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[pm_counter - 2], 2);
			// get crc16 value
			g_stCmdIncomingMager.crc16 = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[pm_counter - 1], 4);
			//iKonkeAfSelfPrint("::seq(%x), crc16(%2x)\r\n", g_stCmdIncomingMager.seq, g_stCmdIncomingMager.crc16);

			// If it is the linkset command, we need to confirm in advance whether CRC verification is required!
			if( g_stCmdIncomingMager.command == ECMD_LINK_SET_REQ ) {
				uint8_t uLinkVerifyEnableSt = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);

				if( uLinkVerifyEnableSt != 0XFF ) {
					g_bLinkCRCVerifyEnableFlag = uLinkVerifyEnableSt;
				}
			}

			// compare the crc16 values!!!
			if((g_bLinkCRCVerifyEnableFlag && (g_stCmdIncomingMager.crc16 == orgl_crc16)) || (!g_bLinkCRCVerifyEnableFlag)) {
				// must be valid command id!
				if( g_stCmdIncomingMager.command != ECMD_UNKNOW ) {
					// Check the validity of parameters in detail according to the pre configuration command information!
					CommandDefSt *pprec = kCmdGetCommandConf(g_stCmdIncomingMager.command);
					if( pprec ) {
						#if 0
						iKonkeAfSelfPrint("PRE-CONFIG: cmd(%d), req(%s), ack(%s), incoming_param_type(%s)\r\n"
							, pprec->command, pprec->cmdname_req, pprec->cmdname_ack, pprec->incoming_param_type);
						#endif
						if((pm_counter - 2) == strlen(pprec->incoming_param_type)) {
							uint16_t item_value = 0, item_length = 0;

							// check params type.
							for(int index = 0; index < (pm_counter - 2); ++index ) {
								if( kUtilsIsOct(g_stCmdIncomingMager.param[index])) {
									//iKonkeAfSelfPrint("ppppm-type-check: type(%c), len(%d)\r\n", pprec->incoming_param_type[index], strlen(g_stCmdIncomingMager.param[index]));
									switch(pprec->incoming_param_type[index]) {
										case ('i'):
											err = (kUtilsIsNumber(g_stCmdIncomingMager.param[index]))?err:KET_ERR_INVALID_PARAM;
											break;
										case ('u'):
											err = (strlen(g_stCmdIncomingMager.param[index]) == 2)?err:KET_ERR_INVALID_PARAM;
											break;
										case ('v'):
											err = (strlen(g_stCmdIncomingMager.param[index]) == 4)?err:KET_ERR_INVALID_PARAM;
											break;
										case ('w'):
											err = (strlen(g_stCmdIncomingMager.param[index]) == 8)?err:KET_ERR_INVALID_PARAM;
											break;
										case ('l'):
											err = (strlen(g_stCmdIncomingMager.param[index]) ==16)?err:KET_ERR_INVALID_PARAM;
											break;
										case ('a'):
											item_value  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[index], 4);
											item_length = strlen(g_stCmdIncomingMager.param[index] + 4);

											if( item_length  % 2 == 0 ) {
												item_length /= 2;

												if( item_value != item_length ) {
													err = KET_ERR_INVALID_PARAM;
												}
											}else {
												err = KET_ERR_INVALID_PARAM;
											}
											break;
										case ('b'):
											item_value  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[index], 4);
											item_length = strlen(g_stCmdIncomingMager.param[index] + 4);

											if( item_length  % 4 == 0 ) {
												item_length /= 4;

												if( item_value != item_length ) {
													err = KET_ERR_INVALID_PARAM;
												}
											}else {
												err = KET_ERR_INVALID_PARAM;
											}
											break;
										case ('c'):
											item_value  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[index], 4);
											item_length = strlen(g_stCmdIncomingMager.param[index] + 4);

											if( item_length % 8 == 0 ) {
												item_length /= 8;

												if( item_value != item_length ) {
													err = KET_ERR_INVALID_PARAM;
												}
											}else {
												err = KET_ERR_INVALID_PARAM;
											}
											break;
										case ('d'):
											item_value  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[index], 4);
											item_length = strlen(g_stCmdIncomingMager.param[index] + 4);

											if( item_length  % 16 == 0 ) {
												item_length /= 16;

												if( item_value != item_length ) {
													err = KET_ERR_INVALID_PARAM;
												}
											}else {
												err = KET_ERR_INVALID_PARAM;
											}
											break;
										default:
											err = KET_ERR_INVALID_PARAM;
											break;
									}
								}else {
									err = KET_ERR_INVALID_PARAM;
								}

								if( err != KET_OK ) break;
							}
						}else {
							err = KET_ERR_CMD_FORMAT;
						}
					}else {
						err = KET_ERR_CMD_INVALID;
					}
				}else {
					err = KET_ERR_CMD_INVALID;
				}
			}else {
				err = KET_ERR_CRC_WRONG;
			}
		}
	}else {
		err = KET_ERR_CMD_FORMAT;
	}

	return err;
}

/*	DESP: sent the link ack for sequence number.
 * 	Auth: dingmz_frc.20191018.
 * */
kk_err_t kCmdLinkAck(LinkStatusEm status )
{
	if( g_bLinkAckEnableFlag ) {
		kk_err_t err = kCmdOutgoingCommandSetUp(ECMD_ULINK_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
		if( err == KET_OK ) {
			kCmdOutgoingCommandPutInU8int(status, true);
			kCmdOutgoingCommandSent();
		}
	}

	return KET_OK;
}

#if 0
if( previousCharacter == ':' && next == ':' ) {
	_ota_data_index = 0;
	while(true ) {
		if( emberSerialReadByte(sizeOrPort, &next) == EMBER_SUCCESS ) {
			// ?D???¡ìo???¨¨??¡ì????¡ìa?¡ìo????¡ì??¡ìo?¡ì1OTA1?¡ì??tPAGE?¡ìoy?Y
			if( next == '\r' || next == '\n' ) {
				g_command_buffer[_ota_data_index] = '\0';
				kkOtaImagePageParser(g_command_buffer);
				// ?¨¤?D???D?3?¡§o???£¤?¨º??¡è??¡§¡ã¡§????¡§????????¡§¡é??¡§?¡§o?
				emberCommandReaderInit();
				// add by dingmz_frc.20181102: clear the input buffer
				_CLEAR_SERIAL_IN_BUFFER_();
				return true;
			}else {
				if( _ota_data_index < sizeof(g_command_buffer)) {
					g_command_buffer[_ota_data_index++] = next;
				}else {
					// ?¨¤?D???D?3?¡§o???£¤?¨º??¡è??¡§¡ã¡§????¡§????????¡§¡é??¡§?¡§o?
					emberCommandReaderInit();
					// add by dingmz_frc.20181102: clear the input buffer
					_CLEAR_SERIAL_IN_BUFFER_();
					return true;
				}
			}
		}
	}
}
#endif

/*	DESP: Obtain the information of child nodes according to the short address.
 * 	Auth: dingmz_frc.20191019.
 * */
void kCmdGetChildDataByNodeID(uint16_t node_id, EmberChildData *childData )
{
	uint8_t index = 0, size = emberAfGetChildTableSize();

	childData->type = EMBER_UNKNOWN_DEVICE;

	for(index = 0; index < size; ++ index ) {
		EmberStatus status = emberAfGetChildData(index, childData);

		if( status == EMBER_SUCCESS ) {
			if( childData->id == node_id ) {
				break;
			}
		}
	}

	if( childData->type == EMBER_UNKNOWN_DEVICE ) {
		memset(childData->eui64, 0, EUI64_SIZE);
		childData->id = 0;
	}
}

/*	DESP: Obtain the information of child nodes according to the IEEE address.
 * 	Auth: dingmz_frc.20191019.
 * */
void kCmdGetChildDataByIEEE(EmberEUI64 ieee, EmberChildData *childData )
{
	uint8_t index = 0, size = emberAfGetChildTableSize();

	childData->type = EMBER_UNKNOWN_DEVICE;

	for(index = 0; index < size; ++ index ) {
		EmberStatus status = emberAfGetChildData(index, childData);

		if( status == EMBER_SUCCESS ) {
			if( memcmp(childData->eui64, ieee, EUI64_SIZE) == 0 ) {
				break;
			}
		}
	}

	if( childData->type == EMBER_UNKNOWN_DEVICE ) {
		memset(childData->eui64, 0, EUI64_SIZE);
		childData->id = 0;
	}
}

/*	DESP: Custom private command processing interface.
 * 	Auth: dingmz_frc.20191019.
 * */
kk_err_t kCmdPrivateCommandHandler(void )
{
	#if 0
	iKonkeAfSelfPrint("Command: name(%s), params(", kCmdGetCommandName(g_stCmdIncomingMager.command));
	int index = 0;
	while(g_stCmdIncomingMager.param[index]) {
		if( index > 0 ) {
			iKonkeAfSelfPrint(":");
		}
		iKonkeAfSelfPrint(g_stCmdIncomingMager.param[index]);
		++index;
	}
	iKonkeAfSelfPrint(")\r\n");
	#endif

	kk_err_t err = KET_OK;

	switch(g_stCmdIncomingMager.command) {
		case (ECMD_LINK_SET_REQ):
		{
			uint8_t u8LinkAckEnable = (bool)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 2);

			if( u8LinkAckEnable != 0XFF ) {
				g_bLinkAckEnableFlag = u8LinkAckEnable;
			}

			err = kCmdOutgoingCommandSetUp(ECMD_LINK_SET_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(0x0, true);
				kCmdOutgoingCommandPutInU8int(g_bLinkAckEnableFlag, true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
		case (ECMD_LOCAL_INFO_REQ):
		{
			EmberEUI64 localEui64 = { 0 };
			EmberNodeType nodeTypeResult = 0xFF;
			EmberNetworkParameters networkParams = { 0 };

			emberAfGetEui64(localEui64);
			emberAfGetNetworkParameters(&nodeTypeResult, &networkParams);

			err = kCmdOutgoingCommandSetUp(ECMD_LOCAL_INFO_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInInt(PROTOCOL_VERSION_DEF, true);
				kCmdOutgoingCommandPutInU16int(networkParams.panId, true);
				kCmdOutgoingCommandPutInU16int(emberAfGetNodeId(), true);
				kCmdOutgoingCommandPutInU8Array(kUtilsReverseArrayU8int(localEui64, EUI64_SIZE, true), EUI64_SIZE, true);
				kCmdOutgoingCommandPutInInt(networkParams.radioChannel, true);
				kCmdOutgoingCommandPutInInt(networkParams.radioTxPower, true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
//		case (ECMD_NETWORK_UP_REQ):
//		{
//			uint8_t opttype = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 2);
//			uint16_t panid  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[1], 4);
//			int channel = atoi(g_stCmdIncomingMager.param[2]);
//			int power   = atoi(g_stCmdIncomingMager.param[3]);
//			EmberStatus status = EMBER_SUCCESS;
//
//			if( opttype == 0X01 ) { // force network creation based on specified parameters!!!
//
//				if( panid == 0XFFFF ) {
//					panid = NETWORK_UP_PAN_DEFAULT;
//				}
//
//				if( channel < 11 || channel > 26 ) {
//					channel = NETWORK_UP_CHANNEL_DEFAULT;
//				}
//
//				if( power < -10 || power > 20 ) {
//					power = NETWORK_UP_POWER_DEFAULT;
//				}
//
//				iKonkeAfSelfPrint("NetworUp: pan(%2x), channel(%d), power(%d)\r\n", panid, channel, power);
//
//				status = emberAfPluginNetworkCreatorNetworkForm(true, panid, power, channel);
//				if( status == EMBER_SUCCESS ) {
//					kk_err_t err = kCmdOutgoingCommandSetUp(ECMD_NETWORK_UP_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
//					if( err == KET_OK ) {
//						kCmdOutgoingCommandPutInU8int(0x00, true);
//						kCmdOutgoingCommandPutInU16int(panid, true);
//						kCmdOutgoingCommandPutInInt(channel, true);
//						kCmdOutgoingCommandPutInInt(power, true);
//						kCmdOutgoingCommandSent();
//					}
//				}
//			}else { // Automatically create a network
//				status = emberAfPluginNetworkCreatorStart(true);
//				if( status == EMBER_SUCCESS ) {
//					err = KET_ERR_OPRATE_IN_PROGRESS;
//				}
//			}
//
//			if( status != EMBER_SUCCESS ) {
//				EmberNodeType nodeTypeResult = 0xFF;
//				EmberNetworkParameters networkParams = { 0 };
//
//				emberAfGetNetworkParameters(&nodeTypeResult, &networkParams);
//
//				err = kCmdOutgoingCommandSetUp(ECMD_NETWORK_UP_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
//				if( err == KET_OK ) {
//					kCmdOutgoingCommandPutInU8int(status, true);
//					kCmdOutgoingCommandPutInU16int(networkParams.panId, true);
//					kCmdOutgoingCommandPutInInt(networkParams.radioChannel, true);
//					kCmdOutgoingCommandPutInInt(networkParams.radioTxPower, true);
//					kCmdOutgoingCommandSent();
//				}
//			}
//
//			break;
//		}
//		case (ECMD_NETWORK_DOWN_REQ):
//		{
//			if( emberAfNetworkState() == EMBER_JOINED_NETWORK ) {
//				networkLeaveCommand();
//				NWK_ClearRssi();
//				 clear the IAS server records
//				//emAfClearServers();
//				err = KET_ERR_OPRATE_IN_PROGRESS;
//			}else {
//				EmberNodeType nodeTypeResult = 0xFF;
//				EmberNetworkParameters networkParams = { 0 };
//
//				emberAfGetNetworkParameters(&nodeTypeResult, &networkParams);
//
//				err = kCmdOutgoingCommandSetUp(ECMD_NETWORK_DOWN_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
//				if( err == KET_OK ) {
//					kCmdOutgoingCommandPutInU8int(0X00, true);
//					kCmdOutgoingCommandSent();
//				}
//			}
//
//			break;
//		}
//		case (ECMD_PJOIN_ON_REQ):
//		{
//			iKonkeAfSelfPrint("PJoin ON\r\n");
//
//			EmberStatus status = emberAfPluginNetworkCreatorSecurityOpenNetwork();
//
//			err = kCmdOutgoingCommandSetUp(ECMD_PJOIN_ON_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
//			if( err == KET_OK ) {
//				kCmdOutgoingCommandPutInU8int(status, true);
//				kCmdOutgoingCommandSent();
//			}
//
//			break;
//		}
//		case (ECMD_PJOIN_OFF_REQ):
//		{
//			iKonkeAfSelfPrint("PJoin OFF\r\n");
//
//			EmberStatus status = emberAfPluginNetworkCreatorSecurityCloseNetwork();
//
//			err = kCmdOutgoingCommandSetUp(ECMD_PJOIN_OFF_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
//			if( err == KET_OK ) {
//				kCmdOutgoingCommandPutInU8int(status, true);
//				kCmdOutgoingCommandSent();
//			}
//
//			break;
//		}
		case (ECMD_LEAVE_REQ):
		{
			EmberEUI64 DestEui64 = { 0, 0, 0, 0, 0, 0, 0, 0 };
			EmberStatus status = EMBER_SUCCESS;
			uint8_t zdo_seq = ZDPAPP_NEXT_SEQ();

			uint16_t parent_id = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);

			for(int index = 0; index < 8; index++) {
				DestEui64[7 - index] = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[1] + index * 2), 2);
			}

			iKonkeAfSelfPrint("Leave REQ: parent_id(%2X), dest_ieee(%X%X%X%X%X%X%X%X)\r\n", parent_id
				, DestEui64[7], DestEui64[6], DestEui64[5], DestEui64[4]
				, DestEui64[3], DestEui64[2], DestEui64[1], DestEui64[0]);

			uint8_t options = EMBER_ZIGBEE_LEAVE_WITHOUT_REJOIN;

			// remove the IAS device record
			//removeServer(DestEui64);
			// send device leave request
			status = emberLeaveRequest(parent_id, DestEui64, options, EMBER_APS_OPTION_RETRY);

			err = kCmdOutgoingCommandSetUp(ECMD_LEAVE_ACK, parent_id, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdo_seq);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(status, true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
		case (ECMD_BIND_REQ):
		{
			EmberEUI64 SrcEui64 = { 0 }, DestEui64 = { 0 };
			uint8_t zdp_seq = ZDPAPP_NEXT_SEQ();

			uint16_t target = (uint16_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint16_t sourceEndpoint = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint16_t destinationEndpoint = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[2], 2);

			for(int index = 0; index < 8; index++) {
				SrcEui64[7 - index]  = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[3] + index * 2), 2);
			}
			for(int index = 0; index < 8; index++) {
				DestEui64[7 - index] = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[4] + index * 2), 2);
			}

			uint16_t clusterId = (uint16_t )kUtilsOct2Value(g_stCmdIncomingMager.param[5], 4);

			EmberStatus status = emberBindRequest(target, SrcEui64, sourceEndpoint, clusterId, UNICAST_BINDING,
				DestEui64, 0, destinationEndpoint, EMBER_AF_DEFAULT_APS_OPTIONS);

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_BIND_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_ACTIVE_EPS_REQ):
		{
			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t zdp_seq = ZDPAPP_NEXT_SEQ();

			EmberStatus status = emberActiveEndpointsRequest(target, EMBER_APS_OPTION_RETRY);

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_ACTIVE_EPS_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_SIMPLE_DESP_REQ):
		{
			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t endpoint= (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint8_t zdp_seq = ZDPAPP_NEXT_SEQ();

			EmberStatus status = emberSimpleDescriptorRequest(target, endpoint, EMBER_AF_DEFAULT_APS_OPTIONS);

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_SIMPLE_DESP_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_MATCH_DESP_REQ):
		{
			uint16_t target  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint16_t profile = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[1], 4);
			uint16_t input_cluster_list[64] = { 0 }, output_cluster_list[64] = { 0 };
			uint8_t zdp_seq = ZDPAPP_NEXT_SEQ();

			int input_cluster_count  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2], 4);
			int output_cluster_count = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[3], 4);

			// get input cluster list.
			for(int index = 0; index < input_cluster_count; ++index ) {
				input_cluster_list[index]  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2] + 4 + (index << 2), 4);
			}

			// get output cluster list.
			for(int index = 0; index < output_cluster_count; ++index ) {
				output_cluster_list[index] = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[3] + 4 + (index << 2), 4);
			}

			EmberStatus status = matchDescriptorsRequest(target, profile, input_cluster_count, output_cluster_count
				, input_cluster_list, output_cluster_list, EMBER_AF_DEFAULT_APS_OPTIONS);

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_MATCH_DESP_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_NODE_DESP_REQ):
		{
			uint16_t target  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t zdp_seq = ZDPAPP_NEXT_SEQ();

			EmberStatus status = emberNodeDescriptorRequest(target, EMBER_AF_DEFAULT_APS_OPTIONS);

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_MATCH_DESP_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandPutInU8int(0x00, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_IEEE_ADDR_REQ):
		{
			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t zdp_seq = ZDPSTK_NEXT_SEQ();

			#if 0 // Do not use the discovery mode, otherwise the response will not be processed in the ZDO callback[emberAfPreZDOMessageReceivedCallback].
			EmberStatus status = emberAfFindIeeeAddress(target, ServiceDiscoveryDoneCallback);
			#else
			EmberStatus status = emberIeeeAddressRequest(target, false, 0, EMBER_APS_OPTION_RETRY);
			#endif

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_IEEE_ADDR_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU64int(0x0000, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_IP_ADDR_REQ):
		{
			uint8_t zdp_seq = ZDPSTK_NEXT_SEQ();
			EmberEUI64 DestEui64 = { 0 };

			for(int index = 0; index < EUI64_SIZE; index++) {
				DestEui64[7 - index]  = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[0] + index * 2), 2);
			}

			#if 0 // Do not use the discovery mode, otherwise the response will not be processed in the ZDO callback[emberAfPreZDOMessageReceivedCallback].
			EmberStatus status = emberAfFindNodeId(DestEui64, ServiceDiscoveryDoneCallback);
			#else
			EmberStatus status = emberNetworkAddressRequest(DestEui64, false, 0);
			#endif

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_IP_ADDR_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zdp_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_CHILD_RECORD_REQ):
		{
			EmberChildData childData = { 0 };
			EmberEUI64 DestEui64 = { 0 };

			uint8_t opttype = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 2);
			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[1], 4);

			for(int index = 0; index < EUI64_SIZE; index++) {
				DestEui64[7 - index]  = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[2] + index * 2), 2);
			}

			childData.type = 0xff;

			if( opttype == 0x0 ) { 		// get child record by node_id
				kCmdGetChildDataByNodeID(target, &childData);
			}
			else if( opttype == 0x1 ) { // get child record by ieee address
				kCmdGetChildDataByIEEE(DestEui64, &childData);
			}

			err = kCmdOutgoingCommandSetUp(ECMD_CHILD_RECORD_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				if( childData.type == EMBER_UNKNOWN_DEVICE ) {
					kCmdOutgoingCommandPutInU8int(0X01, true);
				}else {
					kCmdOutgoingCommandPutInU8int(0X00, true);
				}
				kCmdOutgoingCommandPutInU8int(childData.type, true);
				kCmdOutgoingCommandPutInU16int(childData.id, true);
				kCmdOutgoingCommandPutInU8Array(kUtilsReverseArrayU8int(childData.eui64, EUI64_SIZE, true), EUI64_SIZE, true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
		case (ECMD_CLUSTER_CMD_REQ):
		{
			uint16_t target   = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t des_ep 	  = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint16_t cluster  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2], 4);
			uint16_t mfg_code = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[3], 4);
			uint8_t direction = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[4], 2);
			uint8_t cmd_type  = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[5], 2);
			uint8_t cls_cmd_id= (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[6], 2);
			uint8_t delivery  = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[8], 2);
			uint8_t zcl_seq   = emberAfNextSequence();

			uint8_t tmpZclBuffer[128] = { 0 }, tmpLength = 1;

			tmpZclBuffer[0] = cmd_type;

			// Manufacturer specific
			if( mfg_code > 0 ) {
				tmpZclBuffer[0] |= BIT(2);
				tmpZclBuffer[tmpLength++] = (mfg_code >> 0) & 0xff;
				tmpZclBuffer[tmpLength++] = (mfg_code >> 8) & 0xff;
			}

			// Direction
			if( direction ) {
				tmpZclBuffer[0] |= BIT(3);
			}

			tmpZclBuffer[tmpLength++] = zcl_seq;
			tmpZclBuffer[tmpLength++] = cls_cmd_id;

			EmberStatus status = EMBER_SUCCESS;

			uint16_t payload_len = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[7], 4);
			// Check whether the length is legal.
			if( tmpLength + payload_len < sizeof(tmpZclBuffer)) {
				// put in the ZCL payload content.
				tmpLength += kUtilsStr2Hex(g_stCmdIncomingMager.param[7] + 4, tmpZclBuffer + tmpLength, sizeof(tmpZclBuffer) - tmpLength);

				extern EmberApsFrame globalApsFrame;
				uint8_t src_ep =  kCmdGetEndpointByCluserID(cluster, direction?CLUSTER_MASK_SERVER:CLUSTER_MASK_CLIENT);

				emAfApsFrameClusterIdSetup(cluster);
				emAfApsFrameEndpointSetup(src_ep, des_ep);
				//cliBufferPrint();
				iKonkeAfSelfPrint("-- ZCL-CMD ------\r\n");
				iKonkeAfSelfPrintBuffer(tmpZclBuffer, tmpLength, true);
				iKonkeAfSelfPrint("\r\n-----------------\r\n");

				// sending method.
				switch(delivery) {
					case (0x01):
					{
						status = emberAfSendBroadcast(target, &globalApsFrame, tmpLength, tmpZclBuffer);
						break;
					}
					case (0x02):
					{
						status = emberAfSendMulticast(target, &globalApsFrame, tmpLength, tmpZclBuffer);
						break;
					}
					default:
					{
						status = (target < EMBER_BROADCAST_ADDRESS)?emberAfSendUnicast(EMBER_OUTGOING_DIRECT, target, &globalApsFrame, tmpLength, tmpZclBuffer):
							emberAfSendBroadcast(target, &globalApsFrame, tmpLength, tmpZclBuffer);
						break;
					}
				}
			}else {
				status = EMBER_MESSAGE_TOO_LONG;
			}

			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_CLUSTER_CMD_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zcl_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(0X0000, true);
					kCmdOutgoingCommandPutInU8int(0X00, true);
					kCmdOutgoingCommandPutInU16int(0XFFFF, true);
					kCmdOutgoingCommandPutInU16int(0X0000, true);
					kCmdOutgoingCommandPutInU8int(0XFF, true);
					kCmdOutgoingCommandPutInU8int(0XFF, true);
					kCmdOutgoingCommandPutInU8int(0XFF, true);
					kCmdOutgoingCommandPutInU16int(0X0000, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_READ_ATTR_REQ):
		{
			uint16_t target  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t des_ep 	 = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint16_t cluster = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2], 4);
			uint8_t c_s 	 = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[3], 2);
			uint16_t mfg_code= (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[4], 4);
			uint16_t attr_num= (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[5], 4);

			EmberStatus status = EMBER_SUCCESS;

			if( target == NODE_ID_LOCAL ) { // local read
				err = kCmdOutgoingCommandSetUp(ECMD_READ_ATTR_ACK, target, des_ep, cluster, 0X01, ZCLAPP_NEXT_SEQ());
				if( err == KET_OK ) {
					// get all attributes values
					g_u8GlobalLengthN1 = 0;

					for(int index = 0; index < attr_num; ++index ) {
						uint8_t value_buffer[128] = { 0 }, data_type = ZCL_UNKNOWN_ATTRIBUTE_TYPE;

						uint16_t attribute_id = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[5] + 4 + (4 * index), 4);

						status = emAfReadAttribute(des_ep, cluster, attribute_id, (c_s?CLUSTER_MASK_SERVER:CLUSTER_MASK_CLIENT)
							, mfg_code, value_buffer, sizeof(value_buffer), &data_type);

						g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = (attribute_id >> 8) & 0xff;
						g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = (attribute_id >> 0) & 0xff;
						g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = status;

						// Check whether the property value is read correctly.
						if( status == EMBER_SUCCESS ) {
							uint16_t data_length = (uint16_t)kUtilsGetDataLength(data_type);

							g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = data_type;

							if( data_length > 0 ) {
								kUtilsReverseArrayU8int(value_buffer, data_length, true);
								g_u8GlobalLengthN1 += data_length;
							}else if((data_type == ZCL_LONG_OCTET_STRING_ATTRIBUTE_TYPE) || (data_type == ZCL_LONG_CHAR_STRING_ATTRIBUTE_TYPE) ||
								(data_type == ZCL_ARRAY_ATTRIBUTE_TYPE) || (data_type == ZCL_STRUCT_ATTRIBUTE_TYPE) ||
								(data_type == ZCL_SET_ATTRIBUTE_TYPE) || (data_type == ZCL_BAG_ATTRIBUTE_TYPE)) {
								data_length = U16MERG(value_buffer[1], value_buffer[0]);

								g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = value_buffer[1];
								g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = value_buffer[0];

								memcpy(g_u8GlobalBufferN1 + g_u8GlobalLengthN1, value_buffer + 2, data_length);
								g_u8GlobalLengthN1 += data_length;
							}else {
								data_length = value_buffer[0];
								g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = value_buffer[0];
								memcpy(g_u8GlobalBufferN1 + g_u8GlobalLengthN1, value_buffer + 1, data_length);
								g_u8GlobalLengthN1 += data_length;
							}
						}
					}

					kCmdOutgoingCommandPutInU8int(0X00, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(des_ep, true);
					kCmdOutgoingCommandPutInU16int(cluster, true);
					kCmdOutgoingCommandPutInU8int(c_s, true);
					kCmdOutgoingCommandPutInU16int(mfg_code, true);
					kCmdOutgoingCommandPutInU16int(g_u8GlobalLengthN1, false);
					kCmdOutgoingCommandPutInU8Array(g_u8GlobalBufferN1, g_u8GlobalLengthN1, true);
					kCmdOutgoingCommandSent();
				}
			}else { // global read
				uint8_t zcl_seq = ZCLAPP_NEXT_SEQ();
				uint8_t src_ep  = kCmdGetEndpointByCluserID(cluster, c_s?CLUSTER_MASK_CLIENT:CLUSTER_MASK_SERVER);

				extern uint8_t appZclBuffer[];
				extern uint16_t appZclBufferLen;
				extern uint16_t mfgSpecificId;

				uint8_t mask = c_s?ZCL_FRAME_CONTROL_CLIENT_TO_SERVER:ZCL_FRAME_CONTROL_SERVER_TO_CLIENT;

				mfgSpecificId = mfg_code;
				zclBufferSetup(ZCL_GLOBAL_COMMAND | mask, cluster, ZCL_READ_ATTRIBUTES_COMMAND_ID);
				mfgSpecificId = EMBER_AF_NULL_MANUFACTURER_CODE;

				for(int index = 0; index < attr_num; ++index ) {
					zclBufferAddWord((uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[5] + 4 + (4 * index), 4));
				}

				emAfApsFrameEndpointSetup(src_ep, des_ep);
				// sending method.
				status = emberAfSendUnicast(EMBER_OUTGOING_DIRECT, target, &globalApsFrame, appZclBufferLen, appZclBuffer);
				if( status == EMBER_SUCCESS ) {
					err = KET_ERR_OPRATE_IN_PROGRESS;
				}else {
					err = kCmdOutgoingCommandSetUp(ECMD_READ_ATTR_ACK, target, des_ep, cluster, 0X01, zcl_seq);
					if( err == KET_OK ) {
						kCmdOutgoingCommandPutInU8int(status, true);
						kCmdOutgoingCommandPutInU16int(target, true);
						kCmdOutgoingCommandPutInU8int(des_ep, true);
						kCmdOutgoingCommandPutInU16int(cluster, true);
						kCmdOutgoingCommandPutInU8int(c_s, true);
						kCmdOutgoingCommandPutInU16int(mfg_code, true);
						kCmdOutgoingCommandPutInU16int(0X0000, true);
						kCmdOutgoingCommandSent();
					}
				}
			}

			break;
		}
		case (ECMD_WRITE_ATTR_REQ):
		{
			uint16_t target  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t des_ep 	 = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint16_t cluster = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[2], 4);
			uint8_t c_s 	 = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[3], 2);
			uint16_t mfg_code= (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[4], 4);

			g_u8GlobalLengthN2 = kUtilsStr2Hex(g_stCmdIncomingMager.param[5] + 4, g_u8GlobalBufferN2, GLOBAL_BUFFER_SIZE);

			EmberStatus status = EMBER_SUCCESS;

			if( target == NODE_ID_LOCAL ) { // local write
				err = kCmdOutgoingCommandSetUp(ECMD_WRITE_ATTR_ACK, target, des_ep, cluster, 0X04, ZCLAPP_NEXT_SEQ());
				if( err == KET_OK ) {
					// get all attributes values
					g_u8GlobalLengthN1 = 0;

					for(int index = 0; index < g_u8GlobalLengthN2; ) {
						uint16_t attribute_id = U16MERG(g_u8GlobalBufferN2[index + 0], g_u8GlobalBufferN2[index + 1]);
						uint8_t data_type = g_u8GlobalBufferN2[index + 2];

						status = emAfWriteAttribute(des_ep, cluster, attribute_id, (c_s?CLUSTER_MASK_SERVER:CLUSTER_MASK_CLIENT)
							, mfg_code, g_u8GlobalBufferN2 + index + 3, data_type, true, false);

						// Check whether the property value is read correctly.
						if( status != EMBER_SUCCESS ) {
							g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = status;
							g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = (attribute_id >> 8) & 0xff;
							g_u8GlobalBufferN1[g_u8GlobalLengthN1++] = (attribute_id >> 0) & 0xff;
						}

						uint16_t data_length = (uint16_t)kUtilsGetDataLength(data_type);
						if( data_length == 0 ) {
							data_length = g_u8GlobalBufferN2[index + 3] + 1;
						}

						index += (3 + data_length);
					}

					kCmdOutgoingCommandPutInU8int(0X00, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU8int(des_ep, true);
					kCmdOutgoingCommandPutInU16int(cluster, true);
					kCmdOutgoingCommandPutInU8int(c_s, true);
					kCmdOutgoingCommandPutInU16int(mfg_code, true);
					kCmdOutgoingCommandPutInU16int(g_u8GlobalLengthN1, false);
					kCmdOutgoingCommandPutInU8Array(g_u8GlobalBufferN1, g_u8GlobalLengthN1, true);
					kCmdOutgoingCommandSent();
				}
			}else { // global write
				uint8_t zcl_seq = ZCLAPP_NEXT_SEQ();
				uint8_t src_ep  = kCmdGetEndpointByCluserID(cluster, c_s?CLUSTER_MASK_CLIENT:CLUSTER_MASK_SERVER);

				extern uint8_t appZclBuffer[];
				extern uint16_t appZclBufferLen;
				extern uint16_t mfgSpecificId;

				uint8_t mask = c_s?ZCL_FRAME_CONTROL_CLIENT_TO_SERVER:ZCL_FRAME_CONTROL_SERVER_TO_CLIENT;

				mfgSpecificId = mfg_code;
				zclBufferSetup(ZCL_GLOBAL_COMMAND | mask, cluster, ZCL_WRITE_ATTRIBUTES_COMMAND_ID);
				mfgSpecificId = EMBER_AF_NULL_MANUFACTURER_CODE;

				// convert to small-end format.
				for(int index = 0; index < g_u8GlobalLengthN2; ) {
					kUtilsReverseArrayU8int(g_u8GlobalBufferN2 + index, 2, true);

					uint16_t data_type   = g_u8GlobalBufferN2[index + 2];
					uint16_t data_length = (uint16_t)kUtilsGetDataLength(data_type);
					if( data_length > 0 ) {
						kUtilsReverseArrayU8int(g_u8GlobalBufferN2 + index + 3, data_length, true);
					}else if((data_type == ZCL_LONG_OCTET_STRING_ATTRIBUTE_TYPE) || (data_type == ZCL_LONG_CHAR_STRING_ATTRIBUTE_TYPE) ||
						(data_type == ZCL_ARRAY_ATTRIBUTE_TYPE) || (data_type == ZCL_STRUCT_ATTRIBUTE_TYPE) ||
						(data_type == ZCL_SET_ATTRIBUTE_TYPE) || (data_type == ZCL_BAG_ATTRIBUTE_TYPE)) {

						data_length = U16MERG(g_u8GlobalBufferN2[index + 3], g_u8GlobalBufferN2[index + 4]) + 2;
						kUtilsReverseArrayU8int(g_u8GlobalBufferN2 + index + 3, 2, true);
					}else {
						data_length = g_u8GlobalBufferN2[index + 3] + 1;
					}

					index += (3 + data_length);
				}

				zclBufferAddBuffer(g_u8GlobalBufferN2, g_u8GlobalLengthN2);
				emAfApsFrameEndpointSetup(src_ep, des_ep);

				// sending method.
				status = emberAfSendUnicast(EMBER_OUTGOING_DIRECT, target, &globalApsFrame, appZclBufferLen, appZclBuffer);
				if( status == EMBER_SUCCESS ) {
					err = KET_ERR_OPRATE_IN_PROGRESS;
				}else {
					err = kCmdOutgoingCommandSetUp(ECMD_WRITE_ATTR_ACK, target, des_ep, cluster, 0X04, zcl_seq);
					if( err == KET_OK ) {
						kCmdOutgoingCommandPutInU8int(status, true);
						kCmdOutgoingCommandPutInU16int(target, true);
						kCmdOutgoingCommandPutInU8int(des_ep, true);
						kCmdOutgoingCommandPutInU16int(cluster, true);
						kCmdOutgoingCommandPutInU8int(c_s, true);
						kCmdOutgoingCommandPutInU16int(mfg_code, true);
						kCmdOutgoingCommandPutInU16int(0X0000, true);
						kCmdOutgoingCommandSent();
					}
				}
			}

			break;
		}
		case (ECMD_SET_IC_REQ):
		{
			EmberEUI64 DestEui64 = { 0, 0, 0, 0, 0, 0, 0, 0 };
			uint8_t u8InstallCode[20] = { 0 };
			EmberStatus status = EMBER_SUCCESS;

			for(int index = 0; index < EUI64_SIZE; index++) {
				DestEui64[7 - index]  = (uint8_t)kUtilsOct2Value((char *)(g_stCmdIncomingMager.param[0] + index * 2), 2);
			}

			uint8_t u8InstallCodeMode = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint8_t u8InstallCodeLength = kUtilsStr2Hex(g_stCmdIncomingMager.param[2] + 4, u8InstallCode, sizeof(u8InstallCode));

			// ΪCRCʽҪCRCУֵӵInstallCodeĩβ
			if( u8InstallCodeMode == 0X0 ) {
				uint16_t crc16 = kUtilsInstallCodeX25CRC16(u8InstallCode, u8InstallCodeLength);
				u8InstallCode[u8InstallCodeLength++] = LOW_BYTE(crc16);
				u8InstallCode[u8InstallCodeLength++] = HIGH_BYTE(crc16);
			}

			iKonkeAfSelfPrint("sInstallCode: ieee_addr(%X%X%X%X%X%X%X%X), mode(%X), install_code("
				, DestEui64[7], DestEui64[6], DestEui64[5], DestEui64[4]
				, DestEui64[3], DestEui64[2], DestEui64[1], DestEui64[0]
				, u8InstallCodeMode
				);

			iKonkeAfSelfPrintBuffer(u8InstallCode, u8InstallCodeLength, false);
			iKonkeAfSelfPrint(")\r\n");

			// Convert the install code to a key.
			EmberKeyData key;

			status = emAfInstallCodeToKey(u8InstallCode, u8InstallCodeLength, &key);
			if( EMBER_SUCCESS == status ) {
				status = emberAddTransientLinkKey(DestEui64, &key);
			}

			err = kCmdOutgoingCommandSetUp(ECMD_SET_IC_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(status, true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[0], true);
				kCmdOutgoingCommandPutInU8int(u8InstallCodeMode, true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[2], true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
		case (ECMD_GET_RSSI_REQ):
		{
			EmberStatus status = EMBER_SUCCESS;
			int8_t rssi_nd = -127;
			uint8_t index_tmp = 0;

			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);

			if(!getEndDeviceRssiByNodeid(target, &rssi_nd, &index_tmp)) {
				status = EMBER_NOT_FOUND;
			}

			err = kCmdOutgoingCommandSetUp(ECMD_GET_RSSI_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(status, true);
				kCmdOutgoingCommandPutInU16int(target, true);
				kCmdOutgoingCommandPutInInt(rssi_nd, true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
		case (ECMD_OOD_CMD_REQ):
		{
			uint16_t target = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint32_t dev_id = (uint32_t)kUtilsOct2Value(g_stCmdIncomingMager.param[1], 8);

			// Integrate complete private-ood data package.
			g_u8GlobalLengthN2 = 0;
			// zcl attribute id
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (OPTTUNNEL_OPTDATA_ID >>  0) & 0xff;
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (OPTTUNNEL_OPTDATA_ID >>  1) & 0xff;
			// zcl attribute type
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = ZCL_CHAR_STRING_ATTRIBUTE_TYPE;
			// zcl attribute value length
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = 0;

			// ood length
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = 0;
			// device id
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (dev_id >> 24) & 0xff;
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (dev_id >> 16) & 0xff;
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (dev_id >>  8) & 0xff;
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (dev_id >>  0) & 0xff;
			// channel
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[2], 2);
			// opcode
			g_u8GlobalBufferN2[g_u8GlobalLengthN2++] = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[3], 2);
			// payload
			g_u8GlobalLengthN2 += kUtilsStr2Hex(g_stCmdIncomingMager.param[4] + 4, g_u8GlobalBufferN2 + g_u8GlobalLengthN2, GLOBAL_BUFFER_SIZE - g_u8GlobalLengthN2);

			// reset attribute value length
			g_u8GlobalBufferN2[3] = g_u8GlobalLengthN2 - 4;
			// reset ood length
			g_u8GlobalBufferN2[4] = g_u8GlobalLengthN2 - 4 - 1;

			extern uint8_t appZclBuffer[];
			extern uint16_t appZclBufferLen;
			extern uint16_t mfgSpecificId;

			mfgSpecificId = EMBER_AF_NULL_MANUFACTURER_CODE;
			zclBufferSetup(ZCL_GLOBAL_COMMAND | ZCL_FRAME_CONTROL_CLIENT_TO_SERVER, PRIV_CLUSTER_FCC0, ZCL_WRITE_ATTRIBUTES_COMMAND_ID);
			zclBufferAddBuffer(g_u8GlobalBufferN2, g_u8GlobalLengthN2);

			uint8_t zcl_seq = ZCLAPP_NEXT_SEQ();

			// sending method.
			EmberStatus status = emberAfSendUnicast(EMBER_OUTGOING_DIRECT, target, &globalApsFrame, appZclBufferLen, appZclBuffer);
			if( status == EMBER_SUCCESS ) {
				err = KET_ERR_OPRATE_IN_PROGRESS;
			}else {
				err = kCmdOutgoingCommandSetUp(ECMD_OOD_CMD_ACK, target, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, zcl_seq);
				if( err == KET_OK ) {
					kCmdOutgoingCommandPutInU8int(status, true);
					kCmdOutgoingCommandPutInU16int(target, true);
					kCmdOutgoingCommandPutInU32int(dev_id, true);
					kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[2], true);
					kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[3], true);
					kCmdOutgoingCommandPutInU16int(0x0000, true);
					kCmdOutgoingCommandSent();
				}
			}

			break;
		}
		case (ECMD_OTA_NOTIFY_REQ):
		{
			EmberAfOtaImageId id;

			uint16_t target    = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);
			uint8_t endpoint   = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);
			uint8_t payload_t  = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[2], 2);
			uint8_t jitter     = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[3], 2);

			id.manufacturerId  = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[4], 4);
			id.imageTypeId 	   = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[5], 4);
			id.firmwareVersion = (uint32_t)kUtilsOct2Value(g_stCmdIncomingMager.param[6], 8);

			bool bSucceed = emberAfOtaServerSendImageNotifyCallback(target, endpoint, payload_t, jitter, &id);

			err = kCmdOutgoingCommandSetUp(ECMD_OTA_NOTIFY_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(bSucceed?0x00:0xff, true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[0], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[1], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[2], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[3], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[4], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[5], true);
				kCmdOutgoingCommandPutInOctetString(g_stCmdIncomingMager.param[6], true);
				kCmdOutgoingCommandSent();
			}

			break;
		}
//		case (ECMD_OTA_BLOCK_ACK):
//		{
//			EmberAfOtaImageId id;
//
//			uint8_t errcode   = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 2);
//			uint8_t offset    = atoi(g_stCmdIncomingMager.param[1]);
//			uint8_t block_size= atoi(g_stCmdIncomingMager.param[2]);
//
//			if( errcode ) {
//				bool bSucessed = kkOtaImageBlockWrite(offset, g_stCmdIncomingMager.param[3]);
//				if( bSucessed == false ) {
//					iKonkeAfSelfPrint("Err: kkOtaImageBlockWrite error!!!\r\n");
//				}
//			}else {
//				iKonkeAfSelfPrint("Err: Block Ack error(%x, %d, %d)\r\n", errcode, offset, block_size);
//			}
//
//			break;
//		}
		case (ECMD_DLINK_ACK):
		{
			uint8_t errcode = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[0], 2);
			uint8_t up_seq  = (uint8_t )kUtilsOct2Value(g_stCmdIncomingMager.param[1], 2);

			// It needs to be re-send when the received link response is CRC16 error(0x81) or the link is busy(0xfe)!!
			if((errcode != ELS_ERR_CRC_WRONG) && (errcode != ELS_ERR_LINK_BUSY)) {
				kCmdRemoveOutgoingCommandRecordsBySeq(up_seq);
			}

			break;
		}
		//add bby maozj 20191220
		case (ECMD_NWK_STEERING_REQ):
		{
			extern void kNwkFactoryReset(bool is_nwk_indicator);
			extern void kNwkJoiningStart(uint32_t u32JoiningDuration, pJoinCompleteCallback callback );
			uint8_t status = 0x00;

			uint16_t duration = (uint16_t)kUtilsOct2Value(g_stCmdIncomingMager.param[0], 4);//unit:S

			if (emberNetworkState() == EMBER_JOINED_NETWORK || emberNetworkState() == EMBER_JOINED_NETWORK ){
				iKonkeAfSelfPrint("#####NWK Is Joined By Usart Duration(0x%2x)\r\n", duration);
				status = 0x01;
			}else {
				kNwkFactoryReset(false);
				//steering timeout can be set
				iKonkeAfSelfPrint("#####Start Steering By Usart Duration(0x%2x)\r\n", duration);
				kNwkJoiningStart(duration * 1000, NULL);
			}

			err = kCmdOutgoingCommandSetUp(ECMD_NWK_STEERING_ACK, NODE_ID_LOCAL, ENDPOINT_UNKNOW, CLUSTER_UNKNOW, CLUSTER_CMD_UNKNOW, TRANSMIT_SEQ_UNKNOW);
			if( err == KET_OK ) {
				kCmdOutgoingCommandPutInU8int(status, true);
				kCmdOutgoingCommandSent();
			}
			break;
		}
		default:
		{
			err = KET_ERR_CMD_INVALID;
			break;
		}
	}

	return err;
}

/*	DESP: Message hook, trying to intercept the private konke customer command message.
 * 	Auth: dingmz_frc.20191018.
 * */
void kkPrivateCommandHook(uint8_t inchar )
{
	static uint8_t s_priv_char = 0;

	// head
	if((s_priv_char == '<') && (inchar == 'k')) {
		g_stCmdIncomingMager.counter = 0;
		g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter++] = '<';
		g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter++] = 'k';
		g_stCmdIncomingMager.brecving = true;
	}
	// tail
	else if((g_stCmdIncomingMager.brecving) && (inchar == '}')) {
		g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter++] = inchar;
		g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter] = '\0';

		//iKonkeAfSelfPrint("-->: %s\r\n", g_stCmdIncomingMager.cmdlins);

		kk_err_t err = kCmdCommandIncomingSegmentSplit();
		if( err == KET_OK ) {
			// Generate corresponding command image record before processing command.
			err = kCmdGenerateIncomingCommandImage();
			if( err == KET_OK ) {
				// first deliver the link succ ack.
				kCmdLinkAck(ELS_SUCCESS);

				err = kCmdPrivateCommandHandler();
				// Note: when KET_ERR_OPRATE_IN_PROGRESS is returned, it means that the command has been processed successfully,
				// but the final response needs to be waited.
				if((err == KET_OK) || (err == KET_ERR_OPRATE_IN_PROGRESS)) {
					// do nothing
				}else {
					iKonkeAfSelfPrint("Err: Command Handle Failed(%d)!!\r\n", err);
				}
			}else if( err == KET_ERR_NO_PERMISSIONS ) {
				kCmdLinkAck(ELS_ERR_LINK_BUSY);
			}else {
				kCmdLinkAck(ELS_ERR_UNKNOW);
			}
		}else if( err == KET_ERR_CMD_FORMAT ) {
			kCmdLinkAck(ELS_ERR_CMD_FORMAT);
		}else if( err == KET_ERR_CRC_WRONG ) { // crc16 error!!
			kCmdLinkAck(ELS_ERR_CRC_WRONG);
		}else if( err == KET_ERR_CMD_INVALID ) {
			kCmdLinkAck(ELS_ERR_INVALID_CMMAND);
		}else if( err == KET_ERR_INVALID_PARAM ) {
			kCmdLinkAck(LS_ERR_INVALID_PARAM);
		}else { // command format error!!
			kCmdLinkAck(ELS_ERR_UNKNOW);
		}

		emberCommandReaderInit();
		// add by dingmz_frc.20181102: clear the input buffer
		_CLEAR_SERIAL_IN_BUFFER_();
		// clear recving flag, and try to recving again!
		g_stCmdIncomingMager.brecving = false;
	}else if( g_stCmdIncomingMager.brecving ) {
		if( g_stCmdIncomingMager.counter < COMMNAD_MSG_MAXLEN ) {
			g_stCmdIncomingMager.cmdlins[g_stCmdIncomingMager.counter++] = inchar;
		}else {
			emberCommandReaderInit();
			// add by dingmz_frc.20181102: clear the input buffer
			_CLEAR_SERIAL_IN_BUFFER_();
			// clear recving flag, and try to recving again!
			g_stCmdIncomingMager.brecving = false;
			// just reply TOO LONG err
			kCmdLinkAck(ELS_ERR_CMD_TOOLONG);
		}
	}

	s_priv_char = inchar;
}

/*	DESP: If there is a valid command in the retrieval queue, handle it immediately if there is one.
 * 	Auth: dingmz_frc.20191018.
 * */
void kCmdCommandImageHandleEventHandler(void )
{
	emberEventControlSetInactive(kCmdCommandImageHandleEventControl);

	for(int index = 0; index < CMDIMAGE_LIST_DEPTH; ++index ) {
		if( g_stCmdImageList[index].command != ECMD_UNKNOW ) {
			#if 1
			iKonkeAfSelfPrint("I-ITEM[%02d] command(%d), node(%2x), ep(%d), cluster(%2x), cls_cmd_id(%x), trans_seq(%d), counter(%d)\r\n"
				, index
				, g_stCmdImageList[index].command, g_stCmdImageList[index].node_id, g_stCmdImageList[index].endpoint
				, g_stCmdImageList[index].cluster_id, g_stCmdImageList[index].cls_cmd_id
				, g_stCmdImageList[index].transmit_seq, g_stCmdImageList[index].valid_counter);
			#endif
		}

		if( g_stCmdImageList[index].valid_counter > 0 ) {
			if( --g_stCmdImageList[index].valid_counter == 0 ) {
				// If it is an up-request command, it needs to be retransmitted in 1s.
				uint8_t section_id = SECTION_ID(g_stCmdImageList[index].command);

				if((section_id == DCMD_ACK_SECTION_ID) || (section_id == UCMD_REQ_SECTION_ID)) {
					// re-send...
					CMDSTR_SENT_NL("%s{%s}", kCmdGetCommandName(g_stCmdImageList[index].command), g_stCmdImageList[index].cmdpmls);

					if( g_stCmdImageList[index].retry_counter > 0 ) {
						if( --g_stCmdImageList[index].retry_counter == 0 ) {
							memset(&g_stCmdImageList[index], 0, sizeof(CommandImageNodeSt));
						}else {
							g_stCmdImageList[index].valid_counter = UCMD_RETRY_PERIOD / IMAGE_LOOP_NMS;
						}
					}
				}else {
					memset(&g_stCmdImageList[index], 0, sizeof(CommandImageNodeSt));
				}
			}
		}else {
			memset(&g_stCmdImageList[index], 0, sizeof(CommandImageNodeSt));
		}
	}

	emberEventControlSetDelayMS(kCmdCommandImageHandleEventControl, IMAGE_LOOP_NMS);
}

/*	DESP: custom private command console module initialization interface.
 * 	Auth: dingmz_frc.20191022.
 * */
void kCmdCommandConsoleModuleInit(void )
{
	// Start command image records validity loop check event.
	emberEventControlSetDelayMS(kCmdCommandImageHandleEventControl, IMAGE_LOOP_NMS);
}

